#!/usr/bin/perl -w
# Copyright Mark Zealey 2002. Released under the GPL.
use strict;
use MSN;
use IO::Handle;
use Curses;
use Text::Wrap;

my $CONFIG = 'config';
my $MSN_VERSION = '0.5.3';
my $PING_SAMPLE = 60;			# Send a ping every X seconds
my $PING_COUNT = 4;			# Start doing the counting thing if no
					# ping response after X seconds.
my $MAX_BUFFER_LEN = 50;		# Max scrollback buffer length
my $MAX_LINE_LEN = 50;			# Max length of the input history

my $time_hires;
BEGIN {
	eval 'use Time::HiRes qw/time/';
	$time_hires = 1 unless $@
}

my $startup_time = time;		# Time we started
my $session = 0;			# Number of the current session
my @sessions;				# Info about all sessions
my ($prompt, $pspecial);
my $last_msg = '';
my $status;				# What's displayed in the status-bar
my $escape = 0;				# Status of the escape key
my $disconnect = 0;			# Disconnect when this is 1
my $hide_input = 0;			# Hide input from being displayed?
my $last_type = time;			# Contains the last time the user typed
my $last_ping = '??';			# Time taken for the last ping
my $ping_last;				# When we sent our last ping
my $ping_got = 0;			# When we got our last ping
my @lines;				# List of the lines we've read from the
					# input
my $line_ptr = 0;			# Pointer to the elem in @lines we are
					# ref'ing
my $char_ptr = 0;			# Current pos we are on the line
my $replace_on = 0;			# Do we replace or insert stuff?

my (%conf, $win, $in_line, $stat_line, $GREEN, $RED, $WHITE);

# This is the big array, defining all the /command's we can handle. It's fairly
# simple, in each array element, the following info is contained:
#
# element | description
# --------+---------------------------------------------------------------------
#       1 | An array-ref to an array of names this function is called. The first
#         | element in this sub-array is the name of the function (without the
#         | handle_ bit on the front) that will be called when any of the names
#         | in the list is typed
# --------+---------------------------------------------------------------------
#       2 | An array-ref of options. Each option must be preceeded by its
#         | required-ness, ie, a string containing some of:
#         |             REQ (required)
#         |             OPT (optional)
#         |             MUL (multiple)
# --------+---------------------------------------------------------------------
#       3 | A string containing the help text for the functions
my @funcs = (
		[
			[qw/away/],
			[],
			'Change your status to "away"'
		], [
			[qw/hide hidden/],
			[],
			'Change your status to "hidden"'
		], [
			[qw/status/],
			[
				'REQ' => 'status'
			],
			'Change your status to "[1]"'
		], [
			[qw/online/],
			[],
			'Change your status to "online"'
		], [
			[qw/window session/],
			[
				'REQ' => 'session'
			],
			'Switches to session [1]'
		], [
			[qw/msg/],
			[
				'REQ' => 'session',
				'REQ MUL' => 'text'
			],
			'Send [2] to session [1]'
		], [
			[qw/quit exit/],
			[],
			'Quits MSN'
		], [
			[qw/part bye/],
			[
				'OPT' => 'session'
			],
			'Leave [1], or the current one if none is specified'
		], [
			[qw/add/],
			[
				'REQ MUL' => 'user'
			],
			'Attempts to add [1] to your buddies list, so you '.
			'will be informed when they are online etc...'
		], [
			[qw/del/],
			[
				'REQ MUL' => 'user'
			],
			'Attempts to remove [1] from your buddies list'
		], [
			[qw/authorize/],
			[
				'REQ MUL' => 'user'
			],
			'Authorizes [1] to add you to their buddie list'
		], [
			[qw/deauthorize/],
			[
				'REQ MUL' => 'user'
			],
			'Un-authorizes [1] to add you to their buddie list'
		], [
			[qw/block ban/],
			[
				'REQ MUL' => 'user'
			],
			'Attempts to block [1], ie stops them from talking to '.
			'you'
		], [
			[qw/unblock unban/],
			[
				'REQ MUL' => 'user'
			],
			'Attempts to unblock [1], ie lets them talk to you ' .
			'again'
		], [
			[qw/chat/],
			[
				'REQ MUL' => 'users'
			],
			'Attempts to start a chat with the [1]'
		], [
			[qw/clear/],
			[
				'OPT' => 'window number'
			],
			'Clears the screen of window [1] (or the current one '.
			'if none specified'
		], [
			[qw/invite/],
			[
				'REQ MUL' => 'users to invite'
			],
			'Invites [1] to the current chat session'
		], [
			[qw/change_fuser setname setfuser nick/],
			[
				'REQ' => 'new username'
			],
			'Changes your "friendly" username to [1]'
		], [
			[qw/file_get get/],
			[
				'REQ' => 'transfer number'
			],
			'Attempts to get the file that is waiting on [1]'
		], [
			[qw/file_noget noget cancel/],
			[
				'REQ' => 'transfer number'
			],
			'Attempts to cancel [1]'
		], [
			[qw/help/],
			[
				'OPT' => 'command'
			],
			'Displays this help, or help on [1]'
		], [
			[qw/ping/],
			[],
			'Pings yourself to check the connection is still alive'
		], [
			[qw/uptime/],
			[],
			'Displays how long the client has been running for'
		], [
			[qw/version/],
			[],
			'Gives you info about the client'
		]
	);

my %funcs;

# We now map all the possible functions into the keys hash, which conveniantly
# all point at the 'primary' function, ie, the one we want called.. :)
for(@funcs) {
	my $pri = $_->[0][0];

	for my $t (@{$_->[0]}) {
		$funcs{$t} = [$pri, $_]
	}
}

# Restore the screen to original state if we get killed
$SIG{$_} = sub { &quit if $disconnect; $disconnect = 1 } for qw/KILL STOP TERM/,
								qw/INT/;

# Set a signal handler to stop the screen being messed up when something bad
# happens
$SIG{__WARN__} = sub { &out(undef, @_) };
$SIG{__DIE__} = sub { &out(undef, @_); &quit };

&init_screen;

&handle_version;
&out(undef, 'Type /help to display help');
&out(undef, 'Attempting to connect...');

# Open the log file
open F, '>>log' or die "Can't open log: $!";
F->autoflush(1);

# Start the MSN module
my $msn = MSN->new(Debug => \*F) or &quit;

# Register this list of callbacks
my @callbacks = (
			qw/connect message new_mail open_chat disconnected/,
			qw/close_chat user_types join part change_state out/,
			qw/file_invite file_start auth_failed unknown_user/,
			qw/auth_success deauth_success add_success del_success/,
			qw/block_success unblock_success add del ping/,
			qw/state_success fuser_success/,
		);

eval '$msn->add_callback($_, \\&do_' . $_ . ')' for @callbacks;

$conf{idle} = 600;		# Default values
&read_config;
$conf{user} ||= &ask(0, 'Please enter your MSN username');
$conf{pass} ||= &ask(1, 'Please enter your MSN password');

my %opts = (
		User => $conf{user},
		Pass => $conf{pass},
	);

$opts{Host} = $conf{host} if exists $conf{host};
$opts{Port} = $conf{port} if exists $conf{port};
$opts{FUser} = $conf{fuser} if exists $conf{fuser};

# Try to connect!
if(!defined $msn->connect(%opts)) {
	&error;
	&quit
}

# We're connected
&update_status;

while(1) {
	$disconnect && &quit;

	# Got anything to do with MSN?
	my $fh = $msn->do_stuff(undef, 1, \*STDIN);	# Check every second
	&quit unless defined $fh;

	if(my $t = &in) {
		if($t =~ m!^/!) {
			&func($t)
		} else {
			&handle_msg("$session $t")
		}
	} elsif($conf{idle} and $msn->get_state eq 'online') {	# Check for idle
		if(time - $last_type > $conf{idle}) {		# We idled!

			$msn->change_my_state('idle');
			&out(0, "Starting auto-idle after $conf{idle} seconds");
		}
	}

	# Lag-timer
	if(defined $ping_last) {
		my $diff = time - $ping_last;

		# If we didn't get a ping after we sent the last, and we're 1
		# second more than when we sent the last, start incrementing
		# the lag counter till we get one back.. (We might have
		# disconnected)
		if($ping_got < $ping_last and $diff >= $PING_COUNT) {
			if($time_hires) {
				$last_ping = sprintf "%0.3f", $diff
			} else {
				$last_ping = int $diff
			}

			&update_status
		}

		&handle_ping if $diff >= $PING_SAMPLE
	}
}

sub handle_quit { &quit }
sub quit {
	&out(undef, "Disconnecting");
	&handle_uptime;
	$msn->disconnect if defined $msn;
	close F;
	$in_line->erase;
	$in_line->refresh;
	endwin;				# Close down the windowing system
	exit
}

sub func {
	my $cmd = shift;

	$cmd =~ s!^/!!;			# Remove leading / if from cmdline

	return &handle_msg($cmd) if $cmd =~ s/^\s+/$session /;

	my ($c, $s) = $cmd =~ /^(\w+)\s*(.*)$/;
	$c ||= '';

	my @matches = sort &complete($c, [keys %funcs]);
	if(@matches == 1) {			# 1 match
		eval '&handle_' . $funcs{$matches[0]}[0] . '($s)';
		&out(undef, $@) if $@;
		return
	} else {
		&list_commands($c)
	}
}

################################################################################
# Handle bits
################################################################################

sub handle_add		{ $msn->add_contact($_)		for split ' ', shift }
sub handle_del		{ $msn->del_contact($_)		for split ' ', shift }
sub handle_block	{ $msn->block_contact($_)	for split ' ', shift }
sub handle_unblock	{ $msn->unblock_contact($_)	for split ' ', shift }
sub handle_authorize	{ $msn->authorize_contact($_)	for split ' ', shift }
sub handle_deauthorize	{ $msn->deauthorize_contact($_)	for split ' ', shift }

sub handle_status	{ $msn->change_my_state(shift) }
sub handle_away		{ &handle_status('away') }
sub handle_online	{ &handle_status('online') }
sub handle_hide		{ &handle_status('hidden') }

sub handle_change_fuser	{ $msn->change_fuser(shift) }
sub handle_chat		{ $msn->start_ring(@_) }
sub handle_clear	{ &out(shift, '') for 0 .. $LINES }	# Hack!

sub handle_version {
	&out(undef, "jMSN version $MSN_VERSION by Mark Zealey",
							'(mark@zealos.org)');
}

sub handle_uptime {
	my ($sec, $min, $hours, $days);

	$days = time - $startup_time;

	$sec = $days % 60;
	$days = int($days / 60);
	$min = $days % 60;
	$days = int($days / 60);
	$hours = $days % 24;
	$days = int($days / 24);

	&out(undef, "Current uptime: ${days}d ${hours}h ${min}m ${sec}s");
}

sub handle_ping {
	# Don't send more than 1 ping at a time, so the lag counter works
	# properly when we're disconnected..
	return if defined $ping_last and $ping_got < $ping_last;

	$msn->ping;
	$ping_last = time
}

sub handle_part {
	my $str = shift;

	$str ||= $session;

	$str =~ /^\d+$/ or return &out(undef, 'Parameters for bye are invalid');

	&out($str, 'You have left');
	$msn->part_sess($str) or return &error;
}

sub handle_window {
	my $cmd = shift;

	if($cmd =~ /^\d+$/) {
		&change_win($cmd);
	} else {
		&error("Invalid window number...")
	}
}

sub handle_file_get {
	my $cmd = shift;

	return if $cmd =~ /\D/;

	$msn->file_get($cmd, 1) or return;
	&out(undef, "Starting file transfer $cmd");
}

sub handle_file_noget {
	my $cmd = shift;

	return if $cmd =~ /\D/;

	$msn->file_get($cmd, 0) or return;
	&out(undef, "Canceling file transfer $cmd");
}

sub handle_invite {
	my $line = shift;

	my ($user, $sess) = $line =~ /^(\S+)\s*(\d+)?$/ or return
							&error('Invalid line');

	$sess ||= $session;
	
	$msn->invite_users($sess, $user);
}

sub handle_msg {
	my $cmd = shift;

	my ($to, $text) = $cmd =~ /^(\S+)\s*(.*)$/i or return;
	$last_msg = $to;
	if($to =~ /^\d+$/ and $msn->sess_exists($to)) {
		&out($to, '<' . $msn->plain_fuser($msn->{FUser}) . '>', $text);
		$msn->send_msg_swb($to, $text)
	}

	# Else maybe start new swb and msg them.
}

sub handle_help {
	my $cmd = shift || '';

	if($cmd =~ /(\S+)/) {			# Help on a specific command
		$cmd = $1;
		my @matches = &complete($cmd, [keys %funcs]);
		if(@matches == 0) {
			return &out(undef, "Invalid command '$cmd'")
		} elsif(@matches == 1) {		# Help for just this one
			my $help = $funcs{$matches[0]}[1];
			my $str = $help->[2];
			my @opts;

			# Walk through the arguments list and turn it into a
			# string...
			my ($arg, $num);
			for(@{$help->[1]}) {
				if($arg) {
					$num++;

					my ($start, $end);
					my @s = split ' ', $arg;
					if(grep {$_ eq 'OPT'} @s) {
						$start = '[';
						$end = ']'
					} else {
						$start = '<';
						$end = '>'
					}

					$str =~ s/\[$num]/${start}${_}${end}/g;

					if(grep {$_ eq 'MUL'} @s) {
						$end .= '...'
					}

					push @opts, $start . $_ . $end;

					undef $arg
				} else {
					$arg = $_
				}
			}

			&out(undef, "/$matches[0]", @opts);
			&out(undef, $str);
			return
		}
	}

	&list_commands($cmd);
}

################################################################################
# Callbacks
################################################################################

sub do_connect { &out(undef, 'Connected!'); $ping_last = 0 }

sub do_auth_failed { &error('Authorization to the server failed. (Bad password',
					'or MSN module problem)'); &quit }

sub do_unknown_user {
	my ($msn, $user) = @_;
	&out($msn->get_sess_id, "Unknown user $user")
}

sub do_disconnected {
	&error('Disconnected!');
	&quit
}

sub do_new_mail {
	my ($msn, $msg) = @_;
	&out($msn->get_sess_id, "New mail from $msg->{From} <$msg->{FromAddr}>",
					"regarding $msg->{Subject}")
}

sub do_message {
	my ($msn, $msg) = @_;
	&out($msn->get_sess_id, '<' . &get_fuser($msg) . "> $msg->{Body}");
}

sub do_user_types {
	my ($msn, $msg) = @_;
	&out($msn->get_sess_id, &get_fuser($msg), 'is typing');
}

sub do_ping {
	my ($msn, $time) = @_;

	$ping_got = time;

	if($time_hires) {
		$last_ping = sprintf "%0.3f", $time
	} else {
		$last_ping = int $time
	}

	&update_status
}

sub do_open_chat {
	my ($msn, $num) = @_;
	&new_win($num);
	&out($num, 'Chat session started');
}

sub do_close_chat {
	my ($msn, $num) = @_;
	&out($num, 'Chat session ended');
	&destroy_win($num);
}

sub do_join {
	my ($msn, $u) = @_;
	&out($msn->get_sess_id, &get_fuser($u), "($u->{User}) has joined");
}

sub do_part {
	my ($msn, $u) = @_;
	&out($msn->get_sess_id, &get_fuser($u), "($u->{User}) has left")
}

sub list_generic {
	my ($str, $msn, $u) = @_;
	&out(0, &get_fuser($u), "($u->{User}) $str")
}
sub list_generic_add	{ list_generic('has been ' . shift, @_) }

sub do_auth_success	{ list_generic_add('authorized', @_) }
sub do_deauth_success	{ list_generic_add('deauthorized', @_) }
sub do_add_success	{ list_generic_add('added', @_) }
sub do_del_success	{ list_generic_add('removed', @_) }
sub do_block_success	{ list_generic_add('blocked', @_) }
sub do_unblock_success	{ list_generic_add('unblocked', @_) }
sub do_del		{ list_generic('has removed you', @_) }

sub do_add {
	my ($msn, $u) = @_;
	list_generic('requests to be added', @_);
	&out(0, "Type /auth $u->{User} to allow them to add you")
}

sub do_state_success	{ &update_status }
sub do_fuser_success	{ &update_status }

sub do_change_state {
	my ($msn, $user) = @_;

	my $hst = &get_fuser($user);
        &out(0, "$hst ($user->{User}) has changed status to '" .
                                    $msn->get_human_state($user->{User}) . "'");
}

sub do_out {
	my ($msn, $reason) = @_;
	&out(undef, "Quitting: $reason");
	&quit
}

sub do_file_invite {
	my ($msn, $msg) = @_;

	&out($msn->get_sess_id, "$msg->{User} requests you to get file",
					"$msg->{File}, size $msg->{Size}");
	&out($msn->get_sess_id, "Type /get $msg->{Num} to get it");
}

sub do_file_start {
	my ($msn, $msg) = @_;

	# Debug I guess
	&out(0, "Transfer $msg->{Num} started, IP $msg->{Host}:$msg->{Port}");
}

################################################################################
# Useful functions
################################################################################

sub read_config {
	# Try to read the config
	if(open G, $CONFIG) {
		my $config_file;
		while(<G>) {
			chomp;
			s/\s*#.*//g;
			next unless $_;
			$config_file .= "$_,"
		}
		close G;
		eval "\%conf = (\%conf, $config_file)"
	} else {
		&out(undef, "Warning: Can't open $CONFIG: $!")
	}
}

sub get_fuser {
	my $h = shift;
	$msn->plain_fuser($msn->get_user($h->{User})->{FUser})
}

# Returns a command from a list
sub complete {
	my ($word, $words) = @_;

	return @$words unless defined $word;

	# First check complete matches
	for(@$words) {
		return ($_) if lc $_ eq lc $word
	}

	# Then check for partial matches and return
	grep /^$word/i, @$words;
}

# window number, message. win == undef and puts to current window
sub out {
	my $num = shift;
	$num = $session if !defined $num;

	return print F "ACK! Bad window print!\n" if !exists $sessions[$num];
	$sessions[$num]{Active} = 1;

	my @lines = split /\n/, wrap('', ' ' x 8, @_);

	&add_line($num, @lines?@lines:(''));
	$win->refresh;

	&update_status
}

sub new_win {
	my $num = shift;

	if(exists $sessions[$num]) {
		# Just add a blank line on if we have had a previous chat on
		# this window
		&out($num, '')
	} else {
		$sessions[$num] = {
					Screen => [],
					Users => {},
					Active => 0,
					Line => 0,
				}
	}
}

sub destroy_win {
	my $num = shift;
	$sessions[$num]{Active} = 1
}

sub change_win {
	my $num = shift;

	return unless exists $sessions[$num];
	$prompt = "[$num]" if !$pspecial;

	return if $num == $session;

	my $a = $sessions[$num];

	$sessions[$session]{Active} = 0;
	$a->{Active} = 0;
	my $s = $a->{Screen};

	$win->erase;
	my $start = @$s - $LINES;
	my $end = $LINES;
	if($start < 0) {
		$end += $start;
		$start = 0;
	}
	my @t = splice @$s, $start, $end;
	push @$s, @t;
	$win->addstr(join("\n", @t) . "\n");
	$win->refresh;
	$session = $num;
	&update_status
}

# Adds a line to be outputted to a specific buffer
sub add_line {
	my $num = shift;

	push @{$sessions[$num]{Screen}}, @_;

	my $t = @{$sessions[$num]{Screen}} - $MAX_BUFFER_LEN;
	splice @{$sessions[$num]{Screen}}, 0, $t, () if $t > 0;

	if($num == $session) {			# We actually output it
		$win->addstr(join("\n", @_) . "\n");
	}
}

# first param is weather to hide the input, the rest are the message
sub ask {
	my $old_hide_input = $hide_input;
	my $old_prompt = $prompt;
	my @old_lines = @lines;
	my $old_line_ptr = $line_ptr;

	$hide_input = shift;
	$pspecial = 1;
	$prompt = "@_";
	@lines = ('');
	$line_ptr = 0;

	beep;
	&update_prompt;

	my $line;
	1 while not defined($line = &in);

	$hide_input = $old_hide_input;
	$prompt = $old_prompt;
	$pspecial = 0;
	@lines = @old_lines;
	$line_ptr = $old_line_ptr;
	&update_prompt;

	$line
}

sub list_commands {
	my $start = shift;

	$start ||= '';

	my %uniq;

	$uniq{$funcs{$_}[0]}++ for &complete($start, [keys %funcs]);

	return &out(undef, "No matches found for $start") unless keys %uniq;

	&out(undef, 'Matches are:');
	for my $t (map {$funcs{$_}[1]} sort keys %uniq) {
		my @commands = @{$t->[0]};
		if(@commands == 1) {			# 1 command
			&out(undef, "  $commands[0]")
		} else {				# Multiple commands
			&out(undef, "  {@commands}")
		}
	}
}

sub init_screen {
	initscr;		# Curses initalization
	noecho;				# Don't echo the chars we recieve

	$Text::Wrap::columns = $COLS;

	# Init colors
	start_color;
	init_pair 1, COLOR_GREEN,  COLOR_BLACK;
	init_pair 2, COLOR_RED,    COLOR_BLACK;
	init_pair 3, COLOR_WHITE,  COLOR_BLUE;    

	$GREEN = COLOR_PAIR(1);  
	$RED   = COLOR_PAIR(2);
	$WHITE = COLOR_PAIR(3);

	# Main window where all the output goes to
	$win = newwin($LINES - 1, 0, 0, 0);
	$win->scrollok(1);		# It's ok to scroll the window
	$win->refresh;

	# New window for reading input from
	$in_line = newwin(1, 0, $LINES - 1, 0);
	$in_line->nodelay(1);		# Non-blocking reading
	$in_line->keypad(1);		# Enable nice character mappings

	# New window for reading input from
	$stat_line = newwin(1, 0, $LINES - 2, 0);
	$stat_line->attron($WHITE);

	&new_win(0);
	&change_win(0)
}

sub update_prompt {
	$in_line->erase;

	my $p = "$prompt ";
	my $curs = $char_ptr + length $p;

	$lines[$line_ptr] = '' if not defined $lines[$line_ptr];
	$p .= $lines[$line_ptr] unless $hide_input;

	# Make this cleverer..
	if($curs >= $COLS) {
		$p = substr $p, 1 - $COLS + $curs, $curs;
		$curs = $COLS
	}

	$in_line->addstr($p);
	$in_line->move(0, $curs);
	$in_line->refresh
}

sub update_status {
	if(defined $msn) {
		$status = "[$last_ping] " . $msn->plain_fuser($msn->{FUser}) . 
					" ($msn->{User}) - " . $msn->get_state
	} else {					# Initializing...
		$status = 'Initializing';
	}

	my @t;
	for(0 .. $#sessions) {
		$_ == $session and $sessions[$_]{Active} = 0;
		push @t, $_ if $sessions[$_]{Active}
	}

	$status .= ' [' . join(',', @t) . ']' if @t;
	$status .= ' ' x ($COLS - length $status);

	$stat_line->erase;
	$stat_line->addstr($status);
	$stat_line->refresh;
	&update_prompt
}

# Return a full line, or undef if not avaliable (non-blocking)
sub in {
	while(1) {
		my $t = $in_line->getch;
		last if $t eq ERR;

		# Auto-idle stuff
		if(defined $msn and $msn->get_state eq 'idle') {
			$msn->change_my_state('online');
			&out(0, 'Unidleing after being idle for',
					int(time - $last_type), 'seconds')
		}

		$last_type = time;

		if($escape) {				# Alt-command
			$escape = 0;

			if($t =~ /^\d$/) {		# alt-num, change window
				&change_win(($t || 10) - 1)
			} elsif($t eq KEY_BACKSPACE or $t eq "\b") {
				my $i = substr $lines[$line_ptr], 0, $char_ptr;
				$i =~ s/([\w\d]*[^\w\d]*)$//;
				substr $lines[$line_ptr], 0, $char_ptr, $i;
				$char_ptr -= length $1
			}
		} elsif($t eq "\e") {			# Escape
			$escape = 1

		} elsif($t eq KEY_BACKSPACE or $t eq "\b") {
			substr $lines[$line_ptr], --$char_ptr, 1, ''
								if $char_ptr > 0
		} elsif($t eq KEY_DC) {
			substr $lines[$line_ptr], $char_ptr, 1, ''
		} elsif($t eq KEY_UP) {
			$line_ptr = 0 if ++$line_ptr >= @lines;
			$char_ptr = length $lines[$line_ptr]
		} elsif($t eq KEY_DOWN) {
			$line_ptr = 0 if --$line_ptr <= -@lines;
			$char_ptr = length $lines[$line_ptr]
		} elsif($t eq KEY_LEFT) {
			$char_ptr-- if $char_ptr > 0
		} elsif($t eq KEY_RIGHT) {
			$char_ptr++ if $char_ptr < length $lines[$line_ptr]
		} elsif($t eq KEY_HOME) {
			$char_ptr = 0
		} elsif($t eq KEY_END) {
			$char_ptr = length $lines[$line_ptr]
		} elsif($t eq KEY_IC) {
			$replace_on = !$replace_on

		} elsif(length($t) != 1) {		# Special char
			# Do nothing

		} elsif($t eq "\t") {{			# Tab-completion
			last if $pspecial;

			if($lines[$line_ptr] eq '') {
				# Just a <tab>, recall last msg command
				$lines[$line_ptr] = "/msg $last_msg ";
				$char_ptr = length $lines[$line_ptr];
				last
			}

			my $last = $lines[$line_ptr];
			($last) = substr($last, 0, $char_ptr) =~ /(\w*)$/;

			# XXX: tab-completion on different array-objects
			my @matches = sort &complete($last, [keys %funcs]);

			if(@matches == 1) {		# 1 match
				my $rem = substr($matches[0], length $last,
						length $matches[0]) . ' ';
				substr $lines[$line_ptr], $char_ptr, 0, $rem;
				$char_ptr += length $rem
			} elsif(@matches) {		# Lots of matches
				&list_commands($last)
			} else {			# No matches
				&out(undef, "No tab-match found for $last")
			}
		}} elsif($t eq "\n") {
			next if $lines[$line_ptr] eq '';
			$lines[0] = $t = $lines[$line_ptr];

			$char_ptr = 0;
			$line_ptr = 0;

			unshift @lines, '';

			pop @lines if @lines > $MAX_LINE_LEN;
			&update_prompt;

			return $t
		} else {
			substr $lines[$line_ptr], $char_ptr++, 0, $t
		}

		&update_prompt
	}

	undef
}

sub error {
	my $text = "@_";
	$text ||= $msn->{Error};
	&out(undef, "Error: $text")
}
